"""Clifs plugin to edit text files"""

import re
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Dict, List, Optional, Union

from rich.live import Live
from rich.panel import Panel
from rich.progress import Progress, TaskID
from rich.table import Table

from clifs import ClifsPlugin
from clifs.utils_cli import (
    MatchHighlighter,
    get_count_progress,
    get_last_action_progress,
    print_line,
    set_style,
    user_query,
)
from clifs.utils_fs import PathGetterMixin, get_unique_path

IO_ERROR_MESSAGE = set_style(
    "Could not read or modify the following file, check that "
    "it is a text file readable with the chosen encoding "
    "'{encoding}' and you have read/write access:\n{file_path}"
)


class StreamingEditor(ClifsPlugin, PathGetterMixin):
    """
    Edit text files based on regular expressions.
    """

    plugin_summary = "Edit text files based on regular expressions"
    plugin_description: str = (
        plugin_summary
        + ". Runs line by line and gives a preview of the changes by default."
    )
    files2process: List[Path]
    dir_dest: Path
    dryrun: bool
    encoding: str
    lines: str
    pattern: str
    replacement: str
    max_previews: int
    dont_overwrite: bool

    @classmethod
    def init_parser(cls, parser: ArgumentParser) -> None:
        """
        Adding arguments to an argparse parser. Needed for all clifs_plugins.
        """
        # add args from FileGetterMixin to arg parser
        super().init_parser_mixin(parser)

        parser.add_argument(
            "-pt",
            "--pattern",
            default=".*",
            help="Pattern identifying the substring to be replaced. "
            "Supports syntax for `re.sub` from regex module "
            "(https://docs.python.org/3/library/re.html). "
            "Note that e.g. a pattern like '.*[\r\n]+' can be used in combination with "
            "an empty replacement to delete the selected lines. "
            "A pattern like '[\r\n]+' in combination with a replacement like "
            "'\ninsert\nlines\n' can be used to append lines to specific lines.",
        )
        parser.add_argument(
            "-rp",
            "--replacement",
            default="",
            help="String to use as replacement. "
            "You can use \\1 \\2 etc. to refer to matching groups.",
        )
        parser.add_argument(
            "-l",
            "--lines",
            type=str,
            help="Lines to edit. If not given all lines are processed. "
            "Supports ranges by giving two integers separated by a hyphen (e.g.'1-5') "
            "or lists of lines given by comma separated integer (e.g. '3,4,10'). ",
        )
        parser.add_argument(
            "-e",
            "--encoding",
            type=str,
            default="utf-8",
            help="Text file encoding.",
        )
        parser.add_argument(
            "-do",
            "--dont_overwrite",
            action="store_true",
            help="Do not overwrite the input file but create a second file "
            "including the suffix '_edited' next to each input file.",
        )
        parser.add_argument(
            "-mp",
            "--max_previews",
            type=int,
            default=5,
            help="Maximum number of line changes shown in the preview mode. "
            "Set to zero to skip preview mode completely. Only for the brave...",
        )

    def __init__(self, args: Namespace) -> None:
        super().__init__(args)

        self.files2process, _ = self.get_paths()
        self.line_nums = self.parse_line_nums()

        self.highlight_match = MatchHighlighter(pattern=self.pattern)

        self.preview_count = 0

        # define progress
        self.progress: Dict[str, Progress] = {
            "counts": get_count_progress(),
            "overall": get_last_action_progress(),
        }
        self.tasks = self.get_tasks()

        self.progress_table = Table.grid()
        self.progress_table.add_row(
            Panel.fit(
                self.progress["overall"],
                title="Progress",
                border_style="cyan",
                padding=(1, 2),
            ),
            Panel.fit(
                self.progress["counts"],
                title="Counts",
                border_style="bright_black",
                padding=(1, 2),
            ),
        )

    def run(self) -> None:
        self.exit_if_nothing_to_process(self.files2process)
        if self.max_previews > 0:
            print_line(self.console, title="PREVIEW")
            for file in self.files2process:
                try:
                    self.preview_replace(file)
                except (IOError, UnicodeDecodeError):
                    self.console.print(
                        IO_ERROR_MESSAGE.format(encoding=self.encoding, file_path=file)
                    )
                    sys.exit(1)
                if self.preview_count >= self.max_previews:
                    break
            print_line(self.console, title="END OF PREVIEW")
            if not user_query(
                'If you want to apply the edits, give me a "yes" or "y" now!'
            ):
                self.console.print("Will not edit files for now. See you soon.")
                sys.exit(0)
        with Live(
            self.progress_table,
            console=self.console,
            auto_refresh=False,
        ) as live:
            for file in self.files2process:
                try:
                    self.replace(file)
                    self.progress["overall"].update(
                        self.tasks["progress"],
                        last_action=f"edited '{file.name}'",
                    )
                    self.progress["overall"].advance(self.tasks["progress"])
                    self.progress["counts"].advance(self.tasks["files_edited"])
                    live.refresh()
                except (IOError, UnicodeDecodeError):
                    self.console.print(
                        IO_ERROR_MESSAGE.format(encoding=self.encoding, file_path=file)
                    )
                    sys.exit(1)

    def parse_line_nums(self) -> Optional[Union[List[int], range]]:
        line_nums: Optional[Union[List[int], range]]
        try:
            if self.lines is None:
                return None
            if "-" in self.lines:
                range_min, range_max = map(int, self.lines.split("-"))
                line_nums = range(range_min, range_max + 1)
            elif "," in self.lines:
                line_nums = list(map(int, self.lines.split(",")))

            else:
                line_nums = [int(self.lines)]

        except ValueError:
            self.console.print(
                set_style(
                    f"Could not parse line input: '{self.lines}'. "
                    "Expecting line numbers >=0 given as either a single integer, "
                    "comma separated list of integers, or a range given in format "
                    "'min_line-max_line'.",
                    "error",
                )
            )
            sys.exit(1)

        if min(line_nums) < 1:
            self.console.print(
                set_style(
                    f"Line input contains numbers smaller than one: '{self.lines}'. "
                    "Please select lines >=1 only.",
                    "error",
                )
            )
            sys.exit(1)
        return line_nums

    def get_tasks(self) -> Dict[str, TaskID]:
        # define overall progress task
        tasks = {
            "progress": self.progress["overall"].add_task(
                "Editing files: ", total=len(self.files2process), last_action="-"
            ),
        }

        # define counter tasks
        tasks["files_edited"] = self.progress["counts"].add_task(
            "Files edited:", total=None
        )
        tasks["lines_changed_total"] = self.progress["counts"].add_task(
            "Lines modified:", total=None
        )
        return tasks

    def preview_replace(self, input_file: Path) -> None:
        self.console.print(
            f"Changes in file '{input_file.name}': "
            + set_style(f"(at: {input_file.parent})", "bright_black")
        )
        file_change_count = 0
        with input_file.open("r", encoding=self.encoding) as input_fh:
            if self.line_nums is None:
                for line_num, line in enumerate(input_fh, 1):
                    if self.preview_count < self.max_previews:
                        mod_line = re.sub(self.pattern, self.replacement, line)
                        if mod_line != line:
                            self.preview_count += 1
                            file_change_count += 1
                            self.print_line_diff(line, mod_line, line_num)
                    else:
                        break
            else:
                for line_num, line in enumerate(input_fh, 1):
                    if self.preview_count < self.max_previews:
                        if line_num in self.line_nums:
                            mod_line = re.sub(self.pattern, self.replacement, line)
                            if mod_line != line:
                                self.preview_count += 1
                                file_change_count += 1
                                self.print_line_diff(line, mod_line, line_num)
                    else:
                        break
            if file_change_count == 0:
                self.console.print("  ----")

    def print_line_diff(self, line: str, mod_line: str, line_num: int) -> None:
        self.console.print(
            f"  l{line_num} old:", self.highlight_match(line.rstrip("\n"))
        )
        self.console.print(f"  l{line_num} new:", mod_line.rstrip("\n"))
        self.console.print()

    def replace(self, input_file: Path) -> None:
        """
        Replace all occurrences of a regex pattern in text file specified replacement.

        We read/write line by line here to avoid memory issues for large files.

        :param input_file: The path to the input text file.
        :return: None
        """

        temp_output_file = get_unique_path(
            input_file.parent / (input_file.stem + "_edited" + input_file.suffix)
        )

        with input_file.open(
            "r", encoding=self.encoding
        ) as input_fh, temp_output_file.open("w", encoding=self.encoding) as output_fh:
            if self.line_nums is None:
                for line_num, line in enumerate(input_fh, 1):
                    mod_line = re.sub(self.pattern, self.replacement, line)
                    if mod_line != line:
                        self.progress["counts"].advance(
                            self.tasks["lines_changed_total"]
                        )
                    output_fh.write(mod_line)
            else:
                for line_num, line in enumerate(input_fh, 1):
                    if line_num in self.line_nums:
                        mod_line = re.sub(self.pattern, self.replacement, line)
                        if mod_line != line:
                            self.progress["counts"].advance(
                                self.tasks["lines_changed_total"]
                            )
                        output_fh.write(mod_line)
                    else:
                        output_fh.write(line)

        if not self.dont_overwrite:
            input_file.unlink()
            temp_output_file.rename(input_file)



import re
import unittest
from pathlib import Path
from unittest.mock import MagicMock, Mock, call, patch

import pytest

from clifs.__main__ import main
from clifs.plugins.edit import StreamingEditor
from tests.common.utils_testing import parametrize_default_ids


def test_parse_line_numbers():
    sed_mock = Mock()

    # single digit
    sed_mock.lines = "7"
    assert StreamingEditor.parse_line_nums(sed_mock) == [7]
    # list
    sed_mock.lines = "4,2,9,11,333333"
    assert StreamingEditor.parse_line_nums(sed_mock) == [4, 2, 9, 11, 333333]
    # range
    sed_mock.lines = "1-4242"
    assert StreamingEditor.parse_line_nums(sed_mock) == range(1, 4243)

    # zero included list
    sed_mock.lines = "0,4,2,9,11,333333"
    with pytest.raises(SystemExit) as e:
        StreamingEditor.parse_line_nums(sed_mock)
    assert e.type == SystemExit
    assert e.value.code == 1

    # zero included range
    sed_mock.lines = "0-111"
    with pytest.raises(SystemExit) as e:
        StreamingEditor.parse_line_nums(sed_mock)
    assert e.type == SystemExit
    assert e.value.code == 1

    # negatives included
    sed_mock.lines = "-5,10"
    with pytest.raises(SystemExit) as e:
        StreamingEditor.parse_line_nums(sed_mock)
    assert e.type == SystemExit
    assert e.value.code == 1

    # improper formatting
    sed_mock.lines = "2-10-13"
    with pytest.raises(SystemExit) as e:
        StreamingEditor.parse_line_nums(sed_mock)
    assert e.type == SystemExit
    assert e.value.code == 1


@parametrize_default_ids(
    ("pattern", "replacement"),
    [("next", "other"), (r"^.*$", "blub")],
)
@parametrize_default_ids("line_nums", [None, [1], [2, 3], [5], range(1, 2)])
def test_replace(pattern, replacement, line_nums):
    sed_mock = MagicMock(
        encoding="utf-8", line_nums=line_nums, pattern=pattern, replacement=replacement
    )

    file_content = [
        "l1 some example line\n",
        "l2 some next example line\n",
        "l3 some next next example line\n",
    ]

    with patch(
        "pathlib.Path.open",
        unittest.mock.mock_open(read_data="".join(file_content)),
    ) as m:
        StreamingEditor.replace(sed_mock, Path("some_file.txt"))

    # check files are properly accessed
    assert m.mock_calls.count(call("r", encoding="utf-8")) == 1
    assert m.mock_calls.count(call("w", encoding="utf-8")) == 1

    # check lines are properly modified and written
    line_nums = range(1, len(file_content) + 1) if line_nums is None else line_nums
    exp_calls = []
    for i, line in enumerate(file_content, 1):
        if i in line_nums:
            exp_calls.append(call(re.sub(pattern, replacement, line)))
        else:
            exp_calls.append(call(line))
    handle = m()
    handle.write.assert_has_calls(exp_calls)


@parametrize_default_ids(
    ("pattern", "replacement"),
    [("next", "other"), (r"^.*$", "blub")],
)
@parametrize_default_ids("line_nums", [None, [1], [2, 3], [5], range(1, 2)])
@parametrize_default_ids("max_previews", [0, 2, 1000])
def test_preview_replace(capsys, pattern, replacement, line_nums, max_previews):
    file_content = [
        "l1 some example line\n",
        "l2 some next example line\n",
        "l3 some next next example line\n",
    ]

    mock_args = MagicMock(
        encoding="utf-8",
        pattern=pattern,
        replacement=replacement,
        max_previews=max_previews,
    )

    with patch(
        "pathlib.Path.open",
        unittest.mock.mock_open(read_data="".join(file_content)),
    ) as m, patch.object(
        StreamingEditor, "get_paths", return_value=["bla", "blub"]
    ), patch.object(StreamingEditor, "parse_line_nums", return_value=line_nums):
        sed = StreamingEditor(mock_args)
        StreamingEditor.preview_replace(sed, Path("some_file.txt"))

    # check files are properly accessed
    assert m.mock_calls.count(call("r", encoding="utf-8")) == 1  # we read
    assert m.mock_calls.count(call("w", encoding="utf-8")) == 0  # we do not write

    captured = capsys.readouterr()
    print(f"{captured.out=}")

    # check lines are properly modified and written
    line_nums = range(1, len(file_content) + 1) if line_nums is None else line_nums
    count_changes = 0
    for i, line in enumerate(file_content, 1):
        mod_line = re.sub(pattern, replacement, line)

        if i in line_nums and line != mod_line and count_changes < max_previews:
            assert f"  l{i} old: {line}" in captured.out
            assert f"  l{i} new: {mod_line}" in captured.out
            count_changes += 1
        else:
            assert f"  l{i} old: {line}" not in captured.out
            assert (
                f"  l{i} new: {re.sub(pattern, replacement, line)}" not in captured.out
            )


@parametrize_default_ids(
    ("pattern", "replacement"),
    [("Lore", "ipsum"), ("\\n", "-")],
)
@parametrize_default_ids("dont_overwrite", [False, True])
@parametrize_default_ids("max_previews", [0, 5, 1000])
def test_streaming_editor(
    dirs_source, pattern, replacement, dont_overwrite, max_previews
):
    # run the actual function to test

    file_contents_initial = {}
    for file in dirs_source[0].rglob("*"):
        if file.is_file():
            with file.open("r") as f:
                file_contents_initial[file.name] = f.read()

    patch_args = [
        "clifs",
        "sed",
        str(dirs_source[0]),
        "--pattern",
        pattern,
        "--replacement",
        replacement,
        "--max_previews",
        str(max_previews),
        "--recursive",
    ]
    if dont_overwrite:
        patch_args.append("--dont_overwrite")

    with patch("sys.argv", patch_args), patch("builtins.input", return_value="yes"):
        main()

    print("Editing went through.")

    files_after_editing = [f.name for f in dirs_source[0].rglob("*") if f.is_file()]
    # check that no files got lost
    for file in file_contents_initial.keys():
        assert file in files_after_editing

    if dont_overwrite:
        for file in dirs_source[0].rglob("*"):
            if file.is_file():
                if "_edited" in file.stem:
                    # check edited files are modified
                    with file.open("r") as f:
                        file_content_modified = f.read()
                    assert file_content_modified == re.sub(
                        pattern,
                        replacement,
                        file_contents_initial[file.name.replace("_edited", "")],
                    )
                else:
                    # check initial files are untouched
                    with file.open("r") as f:
                        file_content_modified = f.read()
                    assert file_content_modified == file_contents_initial[file.name]

    else:
        for file in dirs_source[0].rglob("*"):
            if file.is_file():
                # check that temp files are renamed to original file name
                assert "_edited" not in file.stem
                with file.open("r") as f:
                    file_content_modified = f.read()
                assert file_content_modified == re.sub(
                    pattern, replacement, file_contents_initial[file.name]
                )



"""Clifs plugin for regex-based renaming of files and folders"""

import re
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Counter, List, Literal, Set

from rich.text import Text

from clifs import ClifsPlugin
from clifs.utils_cli import MatchHighlighter, cli_bar, print_line, set_style, user_query
from clifs.utils_fs import INDENT, PathGetterMixin, get_unique_path


class Renamer(ClifsPlugin, PathGetterMixin):
    """
    Rename files or folders based on regular expressions.
    """

    plugin_summary: str = "Rename files or directories using regular expressions"
    plugin_description: str = (
        plugin_summary
        + ". By default a preview mode is running to prevent unpleasant surprises."
    )
    pattern: str
    replacement: str
    rename_dirs: bool
    skip_preview: bool

    @classmethod
    def init_parser(cls, parser: ArgumentParser) -> None:
        """
        Adding arguments to an argparse parser. Needed for all clifs_plugins.
        """
        # add args from FileGetterMixin to arg parser
        super().init_parser_mixin(parser)

        parser.add_argument(
            "-pt",
            "--pattern",
            default=".*",
            help="Pattern identifying the substring to be replaced. "
            "Supports syntax for `re.sub` from regex module "
            "(https://docs.python.org/3/library/re.html).",
        )
        parser.add_argument(
            "-rp",
            "--replacement",
            default="",
            help="String to use as replacement. "
            "You can use \\1 \\2 etc. to refer to matching groups. "
            "E.g. a pattern like '(.+)\\.(.+)' in combination "
            "with a replacement like '\\1_suffix.\\2' will append suffixes.",
        )
        parser.add_argument(
            "-d",
            "--dirs",
            dest="rename_dirs",
            action="store_true",
            help="Rename directories instead of files.",
        )
        parser.add_argument(
            "-sp",
            "--skip_preview",
            action="store_true",
            help="Skip preview on what would happen and rename right away. "
            "Only for the brave...",
        )

    def __init__(self, args: Namespace) -> None:
        super().__init__(args)

        self.counter: Counter[str] = Counter()
        self.files, self.dirs = self.get_paths()
        if self.rename_dirs:
            # process deeper folders first to avoid changing paths on the fly
            self.dirs = self.sort_paths(self.dirs)
        self.highlight_match = MatchHighlighter(pattern=self.pattern)

    def run(self) -> None:
        if not self.rename_dirs:
            if not self.files:
                self.console.print("No files to process")
            else:
                if not self.skip_preview:
                    self.rename(self.files, path_type="files", preview_mode=True)
                    if not user_query(
                        'If you want to apply renaming, give me a "yes" or "y" now!'
                    ):
                        self.console.print("Will not rename for now. See you soon.")
                        sys.exit(0)
                self.rename(self.files, path_type="files", preview_mode=False)
        else:
            if not self.dirs:
                self.console.print("No dirs to process")
            else:
                if not self.skip_preview:
                    self.rename(self.dirs, path_type="dirs", preview_mode=True)
                    if not user_query(
                        'If you want to apply renaming, give me a "yes" or "y" now!'
                    ):
                        self.console.print("Will not rename for now. See you soon.")
                        sys.exit(0)
                self.rename(self.dirs, path_type="dirs", preview_mode=False)

    def rename(
        self,
        paths: List[Path],
        path_type: Literal["files", "dirs"],
        preview_mode: bool = True,
    ) -> None:
        self.counter.clear()
        self.counter["paths_total"] = len(paths)

        self.console.print(f"Renaming {self.counter['paths_total']} {path_type}.")
        paths_to_be_added: Set[Path] = set()
        paths_to_be_deleted: Set[Path] = set()
        if preview_mode:
            print_line(self.console, "PREVIEW")

        num_path = 0
        for num_path, path in enumerate(paths, 1):
            name_old = path.name
            name_new = re.sub(self.pattern, self.replacement, name_old)
            messages: List[Text] = []

            # skip items if renaming would result in bad characters
            found_bad_chars = self.find_bad_char(name_new)
            if found_bad_chars:
                messages.append(
                    Text(
                        f"{INDENT}Error: not doing renaming as it would result "
                        f"in invalid characters: '{','.join(found_bad_chars)}'",
                        style="error",
                    )
                )
                self.counter["bad_results"] += 1
                self.print_rename_message(
                    name_old,
                    name_new,
                    num_path,
                    add_messages=messages,
                    preview_mode=preview_mode,
                )
                continue

            # make sure resulting paths are unique
            path_new = path.parent / name_new
            path_unique = get_unique_path(
                path_new,
                set_taken=paths_to_be_added,
                set_free=paths_to_be_deleted | {path},
            )

            if path_new != path_unique:
                path_new = path_unique
                name_new = path_unique.name
                messages.append(
                    Text(
                        f"{INDENT}Warning: name already exists. Adding number suffix.",
                        style="warning",
                    )
                )
                self.counter["name_conflicts"] += 1

            self.print_rename_message(
                name_old,
                name_new,
                num_path,
                add_messages=messages,
                preview_mode=preview_mode,
            )
            # skip items that are not renamed
            if path_new == path:
                continue

            if not preview_mode:
                path.rename(path_new)
                self.counter["paths_renamed"] += 1
            else:
                paths_to_be_added.add(path_new)
                if path_new in paths_to_be_deleted:
                    paths_to_be_deleted.remove(path_new)
                paths_to_be_deleted.add(path)

        if self.counter["bad_results"] > 0:
            noun = "item" if self.counter["name_conflicts"] == 1 else "items"
            self.console.print(
                set_style(
                    f"Warning: {self.counter['bad_results']} {noun} not renamed, "
                    "as it would result in invalid characters.",
                    "warning",
                )
            )

        if self.counter["name_conflicts"] > 0:
            noun = "change" if self.counter["name_conflicts"] == 1 else "changes"
            self.console.print(
                set_style(
                    f"Warning: {self.counter['name_conflicts']} {noun} would have "
                    "resulted in name conflicts. "
                    "Added number suffixes to get unique names.",
                    "warning",
                )
            )

        if not preview_mode:
            self.console.print(
                f"Hurray, {num_path} {path_type} have been processed, "
                f"{self.counter['paths_renamed']} have been renamed."
            )
        if preview_mode:
            print_line(self.console, "END OF PREVIEW")

    def print_rename_message(  # pylint: disable=too-many-arguments
        self,
        name_old: str,
        name_new: str,
        num_path: int,
        add_messages: List[Text],
        preview_mode: bool = False,
    ) -> None:
        indent = 2
        padding = 35

        if name_new == name_old:
            print_message = Text(
                f"{name_old:{padding}} -> {name_new:{padding}}", style="bright_black"
            )
        else:
            highlight_old_name = self.highlight_match(name_old)
            highlight_old_name.pad_right(padding - len(name_old))
            print_message = highlight_old_name + Text(f" -> {name_new:{padding}}")

        for add_mes in add_messages:
            add_mes.pad_left(1)
            print_message += add_mes
        print_message.pad_left(indent)

        if preview_mode:
            self.console.print(print_message)
        else:
            cli_bar(
                num_path,
                self.counter["paths_total"],
                suffix=print_message,
                console=self.console,
            )

    @staticmethod
    def find_bad_char(string: str) -> List[str]:
        """Check stings for characters causing problems in Windows file system."""
        bad_chars = r"~“#%&*:<>?/\{|}"
        return [x for x in bad_chars if x in string]



"""Utilities for the command line interface"""

import argparse
from typing import Iterable, Optional, Union

from rich.console import Console, RenderableType
from rich.highlighter import Highlighter
from rich.progress import BarColumn, Progress, TaskProgressColumn, TimeRemainingColumn
from rich.rule import Rule
from rich.text import Text
from rich.theme import Theme

THEME_RICH = Theme(
    {
        "bar.complete": "default",
        "bar.finished": "green",
        "bar.back": "bright_black",
        "progress.percentage": "default",
        "progress.remaining": "bright_black",
        "rule.line": "default",
        "warning": "yellow",
        "error": "red",
        "folder": "yellow",
        "regex_match": "underline bright_cyan",
    },
)
CONSOLE = Console(theme=THEME_RICH, highlight=False)


class MatchHighlighter(Highlighter):  # pylint: disable=too-few-public-methods
    """Applies highlighting from a list of regular expressions."""

    def __init__(self, pattern: str):
        self.pattern = pattern
        self.style = "regex_match"

    def highlight(self, text: Text) -> None:
        """Highlight :class:`rich.text.Text` at match location.

        Args:
            text (~Text): Text to highlighted.

        """
        text.highlight_regex(self.pattern, style=self.style)


def set_style(
    string: str,
    style: str = "red",
) -> str:
    """Set rich style for strings.

    :param string: String
    :param style: Style to set, defaults to "red"
    :return: String wrapped in style markup
    """
    if string.endswith("\\") and not string.endswith("\\\\"):
        string += "\\"
    return f"[{style}]{string}[/{style}]"


class LastActionProgress(Progress):
    """Progress showing the last action in a separate line."""

    def get_renderables(self) -> Iterable[RenderableType]:
        """Get a number of renderables for the progress display."""

        table = self.make_tasks_table(self.tasks)
        yield table
        for task in self.tasks:
            yield (
                f"{task.fields.get('last_action_desc','Last action')}: "
                f"{task.fields.get('last_action', '-')}"
            )


def size2str(size: float, color: str = "cyan") -> str:
    """Format data size in bites to nicely readable units.

    :param size: Input size in bites
    :param color: Output color for rich markup, defaults to "cyan"
    :return: String of data size wrapped in color markup
    """
    if size < 1024**2:
        unit = "KB"
        size = round(size / 1024, 2)
    elif size < 1024**3:
        unit = "MB"
        size = round(size / 1024**2, 2)
    elif size < 1024**4:
        unit = "GB"
        size = round(size / 1024**3, 2)
    elif size < 1024**5:
        unit = "TB"
        size = round(size / 1024**4, 2)
    else:
        unit = "PB"
        size = round(size / 1024**5, 2)
    return f"[{color}]{size:7.2f} {unit}[/{color}]"


def cli_bar(  # pylint: disable=too-many-arguments
    status: int,
    total: int,
    suffix: Union[str, Text] = "",
    print_out: bool = True,
    bar_len: int = 20,
    console: Optional[Console] = None,
) -> str:
    """Create progress bar and either print directly to console or return as string.

    :param status: Number of finished steps
    :param total: Total number of expected steps
    :param suffix: Suffix to add to the bar, defaults to ""
    :param print_out: Whether to print directly or not, defaults to True
    :param bar_len: Number of characters used for the bar, defaults to 20
    :param console: rich.console object to print to, defaults to None
    :return: Progress bar including percent indication and the suffix
    """
    filled_len = int(round(bar_len * status / float(total)))
    percents = round(100.0 * status / float(total), 1)
    proc_bar = "█" * filled_len + "-" * (bar_len - filled_len)
    bar_info = f"|{proc_bar}| {percents:5}%"
    if print_out:
        if console:
            console.print(bar_info, suffix)
        else:
            print(bar_info + " " + str(suffix))
    return bar_info + " " + str(suffix)


def user_query(message: str) -> bool:
    """Run user query.

    :param message: Message to show
    :return: Boolean indicating if the user input was "y" or "yes"
    """
    yes = {"yes", "y"}
    print(message)
    choice = input().lower()
    return choice in yes


def print_line(console: Console = CONSOLE, title: str = "") -> None:
    """Print a line to the console.

    :param console: rich.console to print to, defaults to CONSOLE
    :param title: Title included in the line, defaults to ""
    """
    console.print(Rule(title=title, align="center"))


def get_count_progress() -> Progress:
    """Get instance of a count progress.

    :return: progress
    """
    return Progress(
        "{task.description}",
        "{task.completed}",
    )


def get_last_action_progress() -> LastActionProgress:
    """Get instance of a progress bar displaying the last action in a separate line.

    :return: Progress
    """
    return LastActionProgress(
        "{task.description}",
        BarColumn(),
        TaskProgressColumn(),
        TimeRemainingColumn(),
    )


class ClifsHelpFormatter(argparse.HelpFormatter):
    """Help text formatter for argparse.ArgumentParser

    Hides destination variables or positional args.
    Shows default values with string defaults being quoted."""

    # do not show dest variable for optional args in help
    def _get_default_metavar_for_optional(self, action: argparse.Action) -> str:
        return ""

    # show default values with quoted strings
    def _get_help_string(self, action: argparse.Action) -> str:
        action_help = action.help if action.help else ""
        if "%(default)" not in action_help and action.default is not argparse.SUPPRESS:
            defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
            if action.option_strings or action.nargs in defaulting_nargs:
                def_str = "%(default)s"
                if isinstance(action.default, str):
                    def_str = f"'{def_str}'"
                action_help += f" (default: {def_str})"
        return action_help



"""Utilities for the file system"""

import csv
import re
import sys
from argparse import ArgumentParser
from datetime import datetime
from pathlib import Path
from typing import Any, List, Literal, Optional, Set, Tuple

from dateutil.relativedelta import relativedelta

from clifs.utils_cli import CONSOLE, set_style

INDENT = "    "
TIME_INTERVAL_HELPTEXT = """The time interval can be given in units of:
 seconds ('s'), minutes ('min'), hours ('h'), days ('d'), months ('mon'),
 or years ('a' or 'y'). The default unit is days. Hence e.g. and input of '3 mon' would
 be interpreted as three months ago while an input of '1.5' would be interpreted as one
 and a half days ago."""
CTIME_HELPTEXT = """Be aware that the meaning of 'ctime' depends on the operating
 system. On some systems (like Unix) it is the time of the last metadata change, while
 on others (like Windows), it is the creation time
 (see https://docs.python.org/3/library/stat.html)."""


class PathGetterMixin:
    """
    Get paths from a source directory by different filter methods.
    """

    dir_source: Path
    recursive: bool
    filterlist: Path
    filterlistheader: str
    filterlistsep: str
    filterstring: str
    mtime_stamp_older: Optional[str] = None
    mtime_stamp_newer: Optional[str] = None
    ctime_stamp_older: Optional[str] = None
    ctime_stamp_newer: Optional[str] = None

    @staticmethod
    def init_parser_mixin(parser: ArgumentParser) -> None:
        """
        Adding arguments to an argparse parser. Needed for all clifs_plugins.
        """
        group = parser.add_argument_group("optional arguments - file selection")

        parser.add_argument(
            "dir_source",
            type=Path,
            help="Source directory containing the files/folders to be processed.",
        )
        group.add_argument(
            "-r",
            "--recursive",
            action="store_true",
            help="Search recursively in source directory.",
        )
        group.add_argument(
            "-fl",
            "--filterlist",
            default=None,
            type=Path,
            help="Path to a txt or csv file containing a list of files to process. "
            "In case of a csv, separator and header can be provided additionally via "
            "the parameters `filterlistsep` and `filterlistheader`. "
            "If no header is provided, each line in the file is treated as individual "
            "file or folder name.",
        )
        group.add_argument(
            "-flh",
            "--filterlistheader",
            default=None,
            help="Header of the column to use as filter "
            "from a csv provided as `filterlist`. "
            "If no header is provided, "
            "each line in the file is read as individual item name.",
        )
        group.add_argument(
            "-fls",
            "--filterlistsep",
            default=",",
            help="Separator to use for csv provided as filter list.",
        )
        group.add_argument(
            "-fs",
            "--filterstring",
            default=None,
            help="Substring identifying files/folders to be copied. "
            "Not case sensitive.",
        )
        group.add_argument(
            "-mto",
            "--mtime_stamp_older",
            default=None,
            help="Select only files/folders which were last modified more than the "
            f"given period of time ago. {TIME_INTERVAL_HELPTEXT}",
        )
        group.add_argument(
            "-mtn",
            "--mtime_stamp_newer",
            default=None,
            help="Select only files/folders which were last modified more recently "
            f"than the given period of time ago. {TIME_INTERVAL_HELPTEXT}",
        )
        group.add_argument(
            "-cto",
            "--ctime_stamp_older",
            default=None,
            help="Select only files/folders which were created/changed more than the "
            f"given period of time ago. {TIME_INTERVAL_HELPTEXT} {CTIME_HELPTEXT}",
        )
        group.add_argument(
            "-ctn",
            "--ctime_stamp_newer",
            default=None,
            help="Select only files/folders which were created/changed more recently "
            f"than the given period of time ago. {TIME_INTERVAL_HELPTEXT} "
            f"{CTIME_HELPTEXT}",
        )

    def get_paths(self) -> Tuple[List[Path], List[Path]]:
        """Get file and folder paths depending on set filters

        :return: Lists of file paths and folder paths matching the filters respectively
        """
        # get paths by substring filter
        files, dirs = self._get_paths_by_filterstring(
            self.dir_source, filterstring=self.filterstring, recursive=self.recursive
        )

        # filter by list
        if self.filterlist:
            list_filter = self._list_from_csv()
            files = [i for i in files if i.name in list_filter]
            dirs = [i for i in dirs if i.name in list_filter]

        # filter by mtime
        if self.mtime_stamp_older or self.mtime_stamp_newer:
            files, dirs = self.filter_by_time(
                files,
                dirs,
                "st_mtime",
                delta_th_upper=self.mtime_stamp_older,
                delta_th_lower=self.mtime_stamp_newer,
            )

        # filter by ctime
        if self.ctime_stamp_older or self.ctime_stamp_newer:
            files, dirs = self.filter_by_time(
                files,
                dirs,
                "st_ctime",
                delta_th_upper=self.ctime_stamp_older,
                delta_th_lower=self.ctime_stamp_newer,
            )

        return files, dirs

    def filter_by_time(  # pylint: disable=too-many-arguments
        self,
        files: List[Path],
        dirs: List[Path],
        time_stat: Literal["st_ctime", "st_mtime"],
        delta_th_upper: Optional[str],
        delta_th_lower: Optional[str],
    ) -> Tuple[List[Path], List[Path]]:
        th_upper = (
            None if not delta_th_upper else self._get_time_threshold(delta_th_upper)
        )
        th_lower = (
            None if not delta_th_lower else self._get_time_threshold(delta_th_lower)
        )

        files_filtered = []
        dirs_filtered = []
        for file in files:
            path_time = getattr(file.stat(), time_stat)
            if (not th_upper or path_time <= th_upper) and (
                not th_lower or path_time >= th_lower
            ):
                files_filtered.append(file)
        for folder in dirs:
            path_time = getattr(folder.stat(), time_stat)
            if (not th_upper or path_time <= th_upper) and (
                not th_lower or path_time >= th_lower
            ):
                dirs_filtered.append(folder)
        return files_filtered, dirs_filtered

    @staticmethod
    def exit_if_nothing_to_process(items: List[Any]) -> None:
        """Exit running process if list of files to process is empty"""
        if not items:
            CONSOLE.print("Nothing to process.")
            sys.exit(0)

    @staticmethod
    def sort_paths(paths: List[Path]) -> List[Path]:
        """Sort by inverse depth and str

        :param paths: List of paths to sort
        :return: Sorted list of paths
        """
        return sorted(paths, key=lambda x: (-len(x.parents), str(x)))

    @staticmethod
    def _get_paths_by_filterstring(
        dir_source: Path, filterstring: Optional[str] = None, recursive: bool = False
    ) -> Tuple[List[Path], List[Path]]:
        """Get files by substring filter on the file name.

        :param dir_source: directory to search for files in
        :param filterstring: Substring that must be included in a file name.
            If set to None, files are not filtered by substring. Defaults to None.
        :param recursive: Search recursively, defaults to False
        :return: Lists of file paths and dir paths matching the filter respectively
        """
        pattern_search = f"*{filterstring}*" if filterstring else "*"
        if recursive:
            pattern_search = "**/" + pattern_search
        files = []
        dirs = []
        for path in dir_source.glob(pattern_search):
            if path.is_dir():
                dirs.append(path.resolve())
            else:
                files.append(path.resolve())

        return files, dirs

    def _list_from_csv(self) -> List[str]:
        if not self.filterlistheader:
            res_list = self.filterlist.open().read().splitlines()
        else:
            with self.filterlist.open(newline="") as infile:
                reader = csv.DictReader(infile, delimiter=self.filterlistsep)
                res_list = []
                for row in reader:
                    try:
                        res_list.append(row[self.filterlistheader])
                    except KeyError:
                        CONSOLE.print(
                            set_style(
                                "Provided csv does not contain header "
                                f"'{self.filterlistheader}'. Found headers:\n"
                                f"{list(row.keys())}",
                                "error",
                            )
                        )
                        sys.exit(1)
        return res_list

    @staticmethod
    def _get_time_threshold(time_input: str, now: datetime = datetime.now()) -> float:
        try:
            if "." in time_input:
                raise ValueError()
            if time_input.endswith("s"):  # seconds
                quantity = int(time_input.rstrip("s").rstrip())
                threshold = (now - relativedelta(seconds=quantity)).timestamp()

            elif time_input.endswith("min"):  # minutes
                quantity = int(time_input.rstrip("min").rstrip())
                threshold = (now - relativedelta(minutes=quantity)).timestamp()

            elif time_input.endswith("h"):  # hours
                quantity = int(time_input.rstrip("h").rstrip())
                threshold = (now - relativedelta(hours=quantity)).timestamp()

            elif time_input.endswith("mon"):  # months
                quantity = int(time_input.rstrip("mon").rstrip())
                threshold = (now - relativedelta(months=quantity)).timestamp()

            elif time_input.endswith("a") or time_input.endswith("y"):  # years
                quantity = int(time_input.rstrip("a").rstrip("y").rstrip())
                threshold = (now - relativedelta(years=quantity)).timestamp()

            else:  # days (default unit)
                quantity = int(time_input.rstrip("d").rstrip())
                threshold = (now - relativedelta(days=quantity)).timestamp()

        except ValueError:
            CONSOLE.print(
                set_style(
                    f"Input time has invalid format: '{time_input}'.\n"
                    "Expecting an integer optionally followed by one of the "
                    "following unit identifiers:\n"
                    "'s' (seconds), 'm' (minutes), 'h' (hours), 'd' (days), "
                    "or 'y' (years).",
                    "error",
                )
            )
            sys.exit(1)
        return threshold


def get_unique_path(
    path_candidate: Path,
    set_taken: Optional[Set[Path]] = None,
    set_free: Optional[Set[Path]] = None,
) -> Path:
    """Given a name candidate get a unique file name in a given directory.

    Adds number suffixes in form ' (#)' if file name is already taken.

    :param path_candidate: Candidate for a file path.
    :param set_taken: Optional set of additional paths which are considered as already
        taken, defaults to None
    :param set_free: Optional sets of paths that are considered as not taken even if
        corresponding files exist, defaults to None
    :raises ValueError: If there are common elements in 'set_taken' and 'set_free'
    :return: Unique file path
    """
    if set_taken is None:
        set_taken = set()
    if set_free is None:
        set_free = set()
    if intersect := set_taken.intersection(set_free):
        raise ValueError(
            "Params 'set_taken' and 'set_free' contain common elements: \n"
            f"{intersect=}."
        )

    path_new = path_candidate
    if (path_new.exists() or path_new in set_taken) and (path_new not in set_free):
        name_file = path_new.stem
        count_match = re.match(r".* \((\d+)\)$", name_file)
        if count_match:
            count = int(count_match.group(1)) + 1
            name_file = " ".join(name_file.split(" ")[0:-1])
        else:
            count = 2

        while (path_new.exists() or path_new in set_taken) and (
            path_new not in set_free
        ):
            name_file_new = name_file + f" ({count})"
            path_new = path_candidate.parent / (name_file_new + path_candidate.suffix)
            count += 1
    return path_new



"""Test the path getter mixin class"""

import os
import time
from datetime import datetime
from pathlib import Path
from unittest.mock import patch

from dateutil.relativedelta import relativedelta

from clifs.utils_fs import PathGetterMixin
from tests.common.utils_testing import parametrize_default_ids


@parametrize_default_ids("filter_str", [".txt", "2", ""])
@parametrize_default_ids("recursive", [True, False])
@parametrize_default_ids(
    ["path_filterlist", "header_filterlist", "sep_filterlist"],
    [
        (None, None, None),
        ("path_filterlist_txt", None, None),
        ("path_filterlist_csv", "filter", ","),
        ("path_filterlist_tsv", "filter", "\t"),
    ],
)
def test_path_getter(
    dirs_source,
    recursive,
    filter_str,
    path_filterlist,
    header_filterlist,
    sep_filterlist,
    request,
):
    for dir in dirs_source:
        # run the actual function to test
        path_getter = PathGetterMixin()

        path_getter.dir_source = dir
        path_getter.recursive = recursive
        path_getter.filterlist = (
            path_filterlist
            if path_filterlist is None
            else request.getfixturevalue(path_filterlist)
        )
        path_getter.filterlistheader = header_filterlist
        path_getter.filterlistsep = sep_filterlist
        path_getter.filterstring = filter_str

        files_found, dirs_found = path_getter.get_paths()

        pattern = f"*{filter_str}*" if filter_str else "*"

        if path_filterlist is None:
            if recursive:
                assert files_found == [x for x in dir.rglob(pattern) if not x.is_dir()]
                assert dirs_found == [x for x in dir.rglob(pattern) if x.is_dir()]
            else:
                assert files_found == [x for x in dir.glob(pattern) if not x.is_dir()]
                assert dirs_found == [x for x in dir.glob(pattern) if x.is_dir()]

        else:
            exp_files = ["L1_file_2.txt", "L2_file_1.txt", "L3_file_3.txt"]
            exp_dirs = ["subdir_1"]
            if recursive:
                assert files_found == [
                    x
                    for x in dir.rglob(pattern)
                    if not x.is_dir() and x.name in exp_files
                ]
                assert dirs_found == [
                    x for x in dir.rglob(pattern) if x.is_dir() and x.name in exp_dirs
                ]
            else:
                assert files_found == [
                    x
                    for x in dir.glob(pattern)
                    if not x.is_dir() and x.name in exp_files
                ]
                assert dirs_found == [
                    x for x in dir.glob(pattern) if x.is_dir() and x.name in exp_dirs
                ]


@parametrize_default_ids("now_stamp", [1e9, datetime.now().timestamp()])
@parametrize_default_ids("quantity", [1, 7, 12])
@parametrize_default_ids("unit", ["s", "min", "h", "mon", "a", "y", "d", ""])
def test_time_parsing(quantity, unit, now_stamp):
    now = datetime.fromtimestamp(now_stamp)

    path_getter = PathGetterMixin()
    th_stamp = path_getter._get_time_threshold(str(quantity) + unit, now=now)

    if unit == "s":
        assert th_stamp == now_stamp - quantity
    elif unit == "min":
        assert th_stamp == now_stamp - quantity * 60
    elif unit == "h":
        assert th_stamp == now_stamp - quantity * 60 * 60
    elif unit == "" or unit == "d":
        assert th_stamp == now_stamp - quantity * 60 * 60 * 24
    elif unit == "mon":
        th_year = now.year - quantity // 12
        th_month = now.month - quantity % 12
        if th_month < 1:
            th_month += 12
            th_year -= 1
        assert th_stamp == now.replace(year=th_year, month=th_month).timestamp()
    elif unit == "a" or unit == "y":
        assert th_stamp == now.replace(year=now.year - quantity).timestamp()


DELTA = 15


@parametrize_default_ids("mtime_stamp_older", [DELTA - 10, DELTA + 10])
@parametrize_default_ids("mtime_stamp_newer", [DELTA - 10, DELTA + 10])
@parametrize_default_ids("ctime_stamp_older", [DELTA - 10, DELTA + 10])
@parametrize_default_ids("ctime_stamp_newer", [DELTA - 10, DELTA + 10])
def test_time_filters(
    dirs_source,
    mtime_stamp_older,
    mtime_stamp_newer,
    ctime_stamp_older,
    ctime_stamp_newer,
):
    dir = dirs_source[0]

    # we test for days unit
    unit = "d"
    time_stamp = time.time() - DELTA * 60 * 60 * 24

    with patch.object(os.stat_result, "st_ctime", time_stamp), patch.object(
        os.stat_result, "st_mtime", time_stamp
    ):
        path_getter = PathGetterMixin()

        path_getter.dir_source = dir
        path_getter.recursive = True
        path_getter.filterlist = None
        path_getter.filterlistheader = None
        path_getter.filterlistsep = None
        path_getter.filterstring = None

        path_getter.mtime_stamp_older = str(mtime_stamp_older) + unit
        path_getter.mtime_stamp_newer = str(mtime_stamp_newer) + unit
        path_getter.ctime_stamp_older = str(ctime_stamp_older) + unit
        path_getter.ctime_stamp_newer = str(ctime_stamp_newer) + unit

        files_found, dirs_found = path_getter.get_paths()

        if (
            float(mtime_stamp_older) > DELTA
            or float(mtime_stamp_newer) < DELTA
            or float(ctime_stamp_older) > DELTA
            or float(ctime_stamp_newer) < DELTA
        ):
            assert files_found == []
            assert dirs_found == []
        else:
            assert set(files_found) == set(
                x.resolve() for x in dir.rglob("*") if not x.is_dir()
            )
            assert set(dirs_found) == set(
                x.resolve() for x in dir.rglob("*") if x.is_dir()
            )
